package dev.pfaff.altargui;

import dev.pfaff.altargui.handlers.*;
import dev.pfaff.altargui.input.InputKey;
import dev.pfaff.altargui.input.InputMods;
import dev.pfaff.altargui.internal.Config;
import dev.pfaff.altargui.internal.elements.RootElement;
import dev.pfaff.altargui.util.Hasher;
import dev.pfaff.altargui.util.NonNullLong2WidgetInstanceHashMap;
import dev.pfaff.altargui.util.listenable.EventListenable;
import dev.pfaff.log4truth.Logger;
import it.unimi.dsi.fastutil.longs.LongArrayList;
import it.unimi.dsi.fastutil.shorts.ShortArrayList;
import org.jspecify.annotations.Nullable;

import java.util.List;
import java.util.Objects;
import java.util.function.Function;
import java.util.function.Predicate;

import static dev.pfaff.altargui.GuiGlobals.profiler;
import static dev.pfaff.log4truth.StandardTags.DEBUG;
import static java.util.FormatProcessor.FMT;

public final class GuiContext {
	private static final RootElement USER_ROOT = new RootElement(0);
	static final RootElement INSPECTOR_ROOT = new RootElement("INSPECTO");

	private final NonNullLong2WidgetInstanceHashMap elements = new NonNullLong2WidgetInstanceHashMap(64);
	private final LongArrayList elementsRemovalQueue = new LongArrayList();
	private final WidgetInstanceList orderedBuffer = new WidgetInstanceList(512);
	private final ShortArrayList treeBuffer = new ShortArrayList();

	private int scaledWindowWidth;
	private int scaledWindowHeight;

	private final CanvasImpl canvas = new CanvasImpl();
	private final MountContext mountContext = new MountContext(this.canvas, this.orderedBuffer, this.treeBuffer);
	private final StateContextImpl stateContext = new StateContextImpl();

	public final EventListenable preUpdate = new EventListenable();

	private final Widget widget;
	@Nullable
	private WidgetInstance<?> widgetElement;
	private final Inspector inspector = new Inspector(this, orderedBuffer, treeBuffer, canvas.enabledDebugLayers, canvas);
	@Nullable
	private WidgetInstance<?> inspectorElement;
	private float mouseX;
	private float mouseY;

	public GuiContext(Widget widget) {
		this.widget = widget;
	}

	public void dispose() {
		elements.forEachValue(WidgetInstance::dispose);
		canvas.close();
	}

	/**
	 * @param scaleFactor the scale factor that should be applied to the UI.
	 * @param windowWidth the framebuffer width, unscaled.
	 * @param windowHeight the framebuffer height, unscaled.
	 */
	public void setWindowProps(double scaleFactor, int windowWidth, int windowHeight) {
		this.scaledWindowWidth = (int)Math.round(windowWidth / scaleFactor);
		this.scaledWindowHeight = (int)Math.round(windowHeight / scaleFactor);
		Logger.log(() -> STR."setWindowProps: scaleFactor=\{scaleFactor}, windowWidth=\{windowWidth}, windowHeight=\{windowHeight} -> scaledWindowWidth=\{scaledWindowWidth}, scaledWindowHeight=\{scaledWindowHeight}");
		canvas.setWindowProps(scaleFactor, windowWidth, windowHeight);
	}

	public void updateAndPaint() {
		profiler().push("GuiContext.updateAndPaint");
		this.mountContext.time = System.currentTimeMillis();
		var canvas = this.canvas;
		orderedBuffer.clear();
		treeBuffer.clear();
		canvas.reset();

		preUpdate.notifyListeners();

		{
			profiler().push("build user content");
			var element = widgetElement;
			if (element == null) element = widgetElement = stateContext.getState(USER_ROOT.key(), () -> {}, widget);
			profiler().replace("lay out user content");
			layoutRootElement(USER_ROOT, element);
			profiler().replace("mount user content");
			element.mount(mountContext);
			profiler().pop();
		}

		{
			inspector.inspectableCount = orderedBuffer.size();
			profiler().push("build inspector content");
			var element = inspectorElement;
			if (element == null) element = inspectorElement = stateContext.getState(INSPECTOR_ROOT.key(), () -> {}, inspector);
			profiler().replace("lay out inspector content");
			layoutRootElement(INSPECTOR_ROOT, element);
			profiler().replace("mount inspector content");
			element.mount(mountContext);
			profiler().pop();
		}

		profiler().push("Canvas.draw");
		canvas.draw();
		profiler().pop();

		boolean generation = this.mountContext.generation;
		// prune the element state
		{
			var elementsRemovalQueue = this.elementsRemovalQueue;
			var elements = this.elements;
			elements.forEachValue(element -> {
				if (element.generation != generation) {
					element.dispose();
					elementsRemovalQueue.add(element.key);
				}
			});
			elementsRemovalQueue.forEach(elements::remove);
			elementsRemovalQueue.clear();
		}

		this.mountContext.generation = !generation;
		profiler().pop();
	}

	private void layoutRootElement(RootElement root, WidgetInstance<?> element) {
		element.attachTo(root);
		int scaledWindowWidth = this.scaledWindowWidth;
		int scaledWindowHeight = this.scaledWindowHeight;
		element.setAvailableSize(scaledWindowWidth, scaledWindowHeight);
		element.stencilLayout(this.stateContext);
		element.setPosition(0f, 0f);
		element.traceLayout();
		element.setClippedRect(0f, 0f, scaledWindowWidth, scaledWindowHeight);
		element.scissorLayout();
	}

	public boolean charTyped(int codePoint) {
		return iterateDown(CharTypedHandler.class, handler -> handler.charTyped(codePoint));
	}

	public boolean keyPressed(InputKey keyCode, int scanCode, InputMods modifiers) {
		return iterateDown(KeyPressHandler.class, handler -> handler.keyPressed(keyCode, scanCode, modifiers));
	}

	public boolean keyReleased(InputKey keyCode, int scanCode, InputMods modifiers) {
		return iterateDown(KeyReleaseHandler.class, handler -> handler.keyReleased(keyCode, scanCode, modifiers));
	}

	public boolean mousePressed(double mouseX, double mouseY, int button) {
		return iterateDown(MousePressHandler.class, handler -> handler.mousePressed(mouseX, mouseY, button));
	}

	public boolean mouseReleased(double mouseX, double mouseY, int button) {
		return iterateDown(MouseReleaseHandler.class, handler -> handler.mouseReleased(mouseX, mouseY, button));
	}

	public boolean mouseMoved(double mouseX, double mouseY) {
		this.mouseX = (float) mouseX;
		this.mouseY = (float) mouseY;
		return iterateDown(element -> element.mouseMoveHandler, handler -> handler.mouseMoved(mouseX, mouseY));
	}

	public boolean mouseScrolled(double mouseX, double mouseY, double delta) {
		return iterateDown(element -> element.mouseScrollHandler,
				handler -> handler.mouseScrolled(mouseX, mouseY, delta));
	}

	public boolean mouseDragged(double mouseX, double mouseY, int button, double deltaX, double deltaY) {
		return iterateDown(element -> element.mouseDragHandler,
				handler -> handler.mouseDragged(mouseX, mouseY, button, deltaX, deltaY));
	}

	public float lastMouseX() {
		return this.mouseX;
	}

	public float lastMouseY() {
		return this.mouseY;
	}

	/**
	 * Bottom-up element iteration.
	 */
	private <T> boolean iterateUp(Class<T> clazz, Predicate<T> consumer) {
		var orderedBuffer = this.orderedBuffer;
		var a = orderedBuffer.elements();
		var upr = orderedBuffer.size();
		if (upr > a.length) throw new AssertionError();
		for (int i = 0; i < upr;) {
			var element = a[i++];
			if (clazz.isInstance(element)) {
				if (consumer.test((T)element)) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Top-down element iteration.
	 */
	private <T> boolean iterateDown(Function<WidgetInstance<?>, @Nullable T> filter, Predicate<T> consumer) {
		return this.orderedBuffer.iterateDown(filter, consumer);
	}

	/**
	 * @deprecated {@link #iterateDown(Function, Predicate)} with a getter for a type-checked, nullable "this-ref"
	 * field offers significantly better performance.
	 */
	@Deprecated
	private <T> boolean iterateDown(Class<T> clazz, Predicate<T> consumer) {
		return this.orderedBuffer.iterateDown(clazz, consumer);
	}

	public int renderBufferByteCount() {
		return this.canvas.renderBufferByteCount();
	}

	final class StateContextImpl implements StateContext {
		private final NonNullLong2WidgetInstanceHashMap elements = GuiContext.this.elements;

		private static long computeKey(long parentKey, Key.CallSite callSite, Widget widget) {
			profiler().push("computeKey");
			Hasher hasher = callSite instanceof Key.Global ? new Hasher() : new Hasher(parentKey);
			callSite.hash(hasher);
			hasher.trackIdentity(widget.getClass());
			profiler().pop();
			return hasher.key();
		}

		public <W extends Widget> WidgetInstance<W> getState(long parentKey, Key.CallSite callSite, W widget) {
			profiler().push("GuiContext.getState");
			long key;
			try {
				Objects.requireNonNull(callSite, "Call-site must be non-null");
				Objects.requireNonNull(widget, "Cannot get state of a null widget");
				key = computeKey(parentKey, callSite, widget);
			} catch (Throwable e) {
				throw GuiException.wrap(FMT."Computing key of \{widget}@??? under @%016x\{parentKey}", e);
			}

			WidgetInstance<W> element;

			try {
				if (Config.SANITIZE_KEYS) {
					//noinspection unchecked
					element = (WidgetInstance<W>) this.elements.computeIfAbsent(key, i -> {
						// TODO: confirm that WidgetInstantiationArgs is always either allocated in a thread-local buffer
						//  or not allocated at all.

						var e = Objects.requireNonNull(widget.instantiate(new WidgetInstantiationArgs(GuiContext.this, widget.getClass(), i, callSite)));
						Logger.log(() -> FMT."instantiated element: \{widget}@%016x\{i} -> \{e}", List.of(DEBUG));
						return e;
					});
				} else {
					//noinspection unchecked
					element = (WidgetInstance<W>) this.elements.computeIfAbsent(key, i -> {
						// TODO: confirm that WidgetInstantiationArgs is always either allocated in a thread-local buffer
						//  or not allocated at all.

						var e = Objects.requireNonNull(widget.instantiate(new WidgetInstantiationArgs(GuiContext.this, widget.getClass(), i, null)));
						Logger.log(() -> FMT."instantiated element: \{widget}@%016x\{i} -> \{e}", List.of(DEBUG));
						return e;
					});
				}
//				{
//					element = Objects.requireNonNull(widget.instantiate(new WidgetInstantiationArgs(GuiContext.this, widget.getClass(), key)));
//					Logger.log(() -> FMT."instantiated element: \{widget}@%016x\{key} -> \{element}", List.of(DEBUG));
//				}
				checkForKeyCollision(callSite, widget, element);

				profiler().push("element.update");
				var initialized = element.initialized;
				element.configure(this, widget, !initialized);
				if (!initialized) element.initialized = true;
				var element_ = element;
				Logger.log(() -> FMT."updated element: \{widget}@%016x\{key} -> \{element_}", List.of(DEBUG));
				profiler().pop();

				profiler().pop();
				return element;
			} catch (Throwable e) {
				throw GuiException.wrap(FMT."Getting state of \{widget}@%016x\{key}", e);
			}
		}

		@Override
		public <W extends Widget> WidgetInstance<W> getState(WidgetInstance<?> parent, Key.CallSite callSite, W widget) {
			return getState(parent.key, callSite, widget);
		}
	}

	private static <W extends Widget> void checkForKeyCollision(Key.CallSite callSite, W widget, WidgetInstance<W> element) {
		if (Config.CHECKS && element.widgetClass != widget.getClass()) {
			Logger.log(() -> STR."""
			Key collision (widget class): \{element.widgetClass.getName()} != \{widget.getClass().getName()}
			\{element}
			\{widget}""");
		}
		if (Config.CHECKS && Config.SANITIZE_KEYS && !callSite.equals(element.keyObject)) {
			Logger.log(() -> STR."""
			Key collision (call-site): \{element.keyObject} != \{callSite}
			\{element}
			\{widget}""");
		}
	}

	void ensureElementIsManagedOnMount(WidgetInstance<?> element) {
		this.elements.putIfAbsent(element.key, element);
	}
}
