package dev.pfaff.altargui.async;

import dev.pfaff.altargui.util.TaskException;

import java.util.concurrent.Future;

/**
 * @param <V> the value produced by the task upon completion.
 */
public interface Task<V> extends AutoCloseable {
	V join() throws TaskException, InterruptedException;

	V joinUninterruptibly() throws TaskException;

	/**
	 * If this method returns {@code true}, then {@link #join()} and friends are guaranteed not to block.
	 */
	boolean isCompleted();

	/**
	 * "Closes" the task. If possible, this will cancel the task. If not, the method simply returns.
	 */
	@Override
	void close();

	static <V> Task<V> completedTask(V value) {
		// this nullability warning is wrong
		//noinspection ConstantValue
		if (value == null) {
			//noinspection unchecked
			return CompletedTask.COMPLETED_NULL;
		}
		return new CompletedTask<>(value);
	}

	static <V> Task<V> wrap(Future<V> future) {
		return new FutureAsTask<>(future);
	}

	static <V> Task<V> nonCancellable(Task<V> task) {
		if (task instanceof NonCancellableTask<V>) return task;
		return new NonCancellableTask<>(task);
	}

	static <V> V joinOrCancel(Task<V> task) throws TaskException {
		try (task) {
			return task.join();
		} catch (InterruptedException e) {
			throw new TaskException.Cancelled("Join interrupted", e);
		}
	}

	/**
	 * Spawns a new virtual thread that joins the {@code task} and does not attempt to catch exceptions. If one is
	 * thrown, the default uncaught exception handler for the VM will be invoked.
	 */
	static void uncatch(Task<?> task) {
		Thread.startVirtualThread(() -> {
			try {
				task.join();
			} catch (InterruptedException e) {
				throw new RuntimeException(e);
			}
		});
	}
}
