package dev.pfaff.altargui.debug;

import dev.pfaff.log4truth.format.WriterAppender;
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.BiConsumer;

import static dev.pfaff.altargui.util.ThrowUtil.throwUnchecked;

public final class EventTreeProfiler implements Profiler, AutoCloseable {
	private final ReentrantLock lock = new ReentrantLock();
	private final Long2ObjectOpenHashMap<ThreadInstance> threadInstances = new Long2ObjectOpenHashMap<>();
	private final Path dir;
	private final Renderer renderer;

	public EventTreeProfiler(Path dir, Renderer renderer) {
		this.dir = dir;
		this.renderer = renderer;
		try {
			Files.createDirectories(dir);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public void push(String name) {
	}

	@Override
	public void pop() {
	}

	@Override
	public void replace(String name) {
	}

	@Override
	public boolean isEventEnabled(Class<? extends Event> eventClass) {
		return true;
	}

	@Override
	public void pushEvent(Event event) {
		var thread = Thread.currentThread();
		var threadId = thread.threadId();
		ThreadInstance instance;
		lock.lock();
		try {
			instance = this.threadInstances.computeIfAbsent(threadId, _ -> new ThreadInstance());
		} finally {
			lock.unlock();
		}
		int depth = instance.depth;
		BufferedWriter writer = instance.writer;
		if (writer == null) {
			writer = initThreadWriter(thread, instance);
		}
		try {
			this.renderer.pushEvent(writer, event, depth);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private @NonNull BufferedWriter initThreadWriter(Thread thread, ThreadInstance instance) {
		BufferedWriter writer;
		try {
			writer = Files.newBufferedWriter(this.dir.resolve(thread.getName() + this.renderer.fileNameSuffix()));
			instance.writer = writer;
			this.renderer.start(writer);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
		return writer;
	}

	@Override
	public void popEvent() {
		var threadId = Thread.currentThread().threadId();
		lock.lock();
		ThreadInstance instance;
		try {
			instance = this.threadInstances.get(threadId);
		} finally {
			lock.unlock();
		}
		// implicit null check
		int depth = --instance.depth;
		var writer = instance.writer;
		assert writer != null;
		try {
			this.renderer.popEvent(writer, depth);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public void close() throws IOException {
		var renderer = this.renderer;
		var consumer = new BiConsumer<Long, ThreadInstance>() {
			private @Nullable Throwable exception;

			@Override
			public void accept(Long threadId, ThreadInstance instance) {
				try (var writer = instance.writer) {
					if (writer != null) renderer.finish(writer);
				} catch (Throwable e) {
					if (this.exception != null) {
						this.exception.addSuppressed(e);
					} else {
						this.exception = e;
					}
				}
			}
		};
		this.threadInstances.forEach(consumer);
		if (consumer.exception != null) {
			throw throwUnchecked(consumer.exception);
		}
	}

	public interface Renderer {
		String fileNameSuffix();

		void start(BufferedWriter writer) throws IOException;

		/**
		 * @param depth the zero-based depth of the event.
		 */
		void pushEvent(BufferedWriter writer, Event event, int depth) throws IOException;

		/**
		 * @param depth the zero-based depth of the event.
		 */
		void popEvent(BufferedWriter writer, int depth) throws IOException;

		void finish(BufferedWriter writer) throws IOException;
	}

	public static final class TextRenderer implements Renderer {
		public static final TextRenderer INSTANCE = new TextRenderer();

		private TextRenderer() {
		}

		@Override
		public String fileNameSuffix() {
			return "";
		}

		@Override
		public void start(BufferedWriter writer) {
		}

		@Override
		public void pushEvent(BufferedWriter writer, Event event, int depth) throws IOException {
			writer.append('\n');
			for (int i = 0; i < depth; i++) {
				writer.append("  ");
			}
			writer.append("- ");
			writer.append(event.toString());
		}

		@Override
		public void popEvent(BufferedWriter writer, int depth) {
		}

		@Override
		public void finish(BufferedWriter writer) {
		}
	}

	public static final class HTMLRenderer implements Renderer {
		public static final HTMLRenderer INSTANCE = new HTMLRenderer();
		public static final String HTML_EVENT_OPEN = """
				<div class="event"><label class="label"><input type="checkbox\"""";

		private final AtomicLong nextCheckboxId = new AtomicLong(0);

		private HTMLRenderer() {
		}

		@Override
		public String fileNameSuffix() {
			return ".html";
		}

		@Override
		public void start(BufferedWriter writer) throws IOException {
			writer.append("""
					<style>
					    body {
							overflow: hidden;
							max-height: 100vh;
							margin: 0;
						}

						.chart {
							border-radius: 2px;
							border: 2px solid black;
							font-family: sans-serif;
							overflow: hidden;
							margin: 8px;

							& > .chart-inner {
								overflow: auto;
								max-height: calc(100vh - (2px * 2) - (8px * 2));
							}
						}

						.chart-inner, .event, .event > .children {
							display: flex;
							flex-wrap: nowrap;
							gap: 2px;
						}

						.chart-inner, .event > .children {
							flex-direction: column;
						}

						.event {
							flex-shrink: 0;
							flex-direction: row;

							& > .label {
								& > input {
									display: none;
								}

								& + .children {
									display: none;
								}

								&:has(input:checked) {
									background-color: #e3e3e3;
									font-weight: bold;

									& + .children {
										display: unset;
									}
								}

								background-color: #f2f2f2;
								text-align: center;
								padding: 6px 8px;
							}
						}
					</style>
					<div class="chart"><div class="chart-inner">""");
		}

		@Override
		public void pushEvent(BufferedWriter writer, Event event, int depth) throws IOException {
			if (depth == 0) {
				writer.append(HTML_EVENT_OPEN + ">");
			} else {
				writer.append(HTML_EVENT_OPEN + " checked>");
			}
			event.appendTo(new WriterAppender(writer));
			writer.append("""
					</label><div class="children">""");
		}

		@Override
		public void popEvent(BufferedWriter writer, int depth) throws IOException {
			writer.append("""
						</div>
					</div>""");
		}

		@Override
		public void finish(BufferedWriter writer) throws IOException {
			writer.append("</div></div>");
		}
	}

	private static final class ThreadInstance {
		private int depth;
		private @Nullable BufferedWriter writer;
	}
}
