package dev.pfaff.altargui.render;

import dev.pfaff.altargui.GuiGlobals;
import dev.pfaff.altargui.async.BorrowedThreadTaskExecutor;
import dev.pfaff.altargui.async.Task;
import dev.pfaff.altargui.async.TaskExecutor;
import dev.pfaff.altargui.host.ContextForker;
import dev.pfaff.altargui.util.GLThread;
import dev.pfaff.glass.opengl.GL;

import java.util.Comparator;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.Callable;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.stream.IntStream;

import static dev.pfaff.altargui.GuiGlobals.HOST;
import static dev.pfaff.altargui.internal.Config.GL_CONTEXT_POOL_SIZE;

/**
 * A pool of OpenGL worker threads, each with their own context with sharing enabled with the main context.
 */
public abstract sealed class GLContextPool {
	public static ScopedValue<GL> POOL_ACQUIRED_GL = ScopedValue.newInstance();

	public abstract TaskExecutor acquireExecutor();

	/**
	 * The global context pool.
	 */
	public static final GLContextPool GL_CONTEXT_POOL;

	static {
		var forker = HOST.contextForker();
		if (forker != null) {
			GL_CONTEXT_POOL = new PerCore(forker);
		} else {
			GL_CONTEXT_POOL = new OnMainThread();
		}
	}

	private static final class OnMainThread extends GLContextPool {
		private static final TaskExecutor EXECUTOR = new TaskExecutor() {
			@Override
			public <V> Task<V> submit(Callable<V> callable) {
				return GLThread.executor().submit(() -> {
					return ScopedValue.callWhere(POOL_ACQUIRED_GL, GuiGlobals.GL, callable::call);
				});
			}
		};

		@Override
		public TaskExecutor acquireExecutor() {
			return EXECUTOR;
		}
	}

	private static final class PerCore extends GLContextPool {
		private final List<Worker> threadTasks;

		public PerCore(ContextForker forker) {
			this.threadTasks = IntStream.range(0, GL_CONTEXT_POOL_SIZE)
					.mapToObj(i -> {
						var queue = new LinkedBlockingQueue<Runnable>();
						var thread = Thread.ofPlatform().name(STR."gl-worker-\{i}").start(() -> {
							try (var ctx = forker.forkContext()) {
								ScopedValue.runWhere(POOL_ACQUIRED_GL, ctx.gl(), () -> {
									for (; ; ) {
										Runnable task;
										try {
											task = queue.take();
										} catch (InterruptedException e) {
											throw new RuntimeException(e);
										}
										try {
											task.run();
										} catch (Throwable e) {

										}
									}
								});
							}
						});
						return new Worker(new BorrowedThreadTaskExecutor(thread, runnable -> {
							try {
								queue.put(runnable);
							} catch (InterruptedException ex) {
								throw new RuntimeException(ex);
							}
						}), queue);
					})
					.toList();
		}

		@Override
		public TaskExecutor acquireExecutor() {
			var worker = Objects.requireNonNull(this.threadTasks.stream()
					.unordered()
					.min(Comparator.comparing(w -> w.queue.size()))
					.orElse(null));
			return worker.executor;
		}

		private record Worker(BorrowedThreadTaskExecutor executor, LinkedBlockingQueue<Runnable> queue) {}
	}
}
